\\ Copyright 2019, 2020, 2021, 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Work in progress.
\\ Maybe working, maybe not working.
\\
\\     OEIS_print_STU(vec) =
\\     OEIS_antidiagonals_of_func(len,func, row_start=1,col_start=1,down=0)
\\
\\     OEIS_check_func(anum,func=anum,flags=0)
\\     OEIS_check_vecfunc(anum,func=anum,flags=0)
\\     OEIS_check_rowfunc(anum,func=anum,flags=0)
\\     OEIS_check_predfunc(anum,func=anum,flags=0)
\\     OEIS_check_verbose = 0
\\
\\ maybe
\\ OEIS_print_triangle(rowfunc,n_range,k_start)
\\ OEIS_print_array(Tfunc,n_range,k_range)
\\
\\ Check Function
\\ --------------
\\ 
\\ OEIS_check_func() checks a function against the squence DATA,
\\ 
\\     A016742(n) = (2*n)^2;         \\ function
\\     OEIS_check_func("A016742")    \\ check it
\\     
\\ The function name must start with the A-number but can have extra after
\\ that as for instance a reminder of what it does
\\ 
\\     A000578_cube(n) = n^3;
\\     OEIS_check_func("A000578_cube")
\\ 
\\ An arbitrary function name or closure (anonymous function) can be given,
\\ 
\\     triple(n) = 3*n;
\\     OEIS_check_func("A008585", "triple");  \\ global
\\     my(trip(n)=3*n);
\\     OEIS_check_func("A008585", trip);      \\ closure
\\     OEIS_check_func("A008585", (n)->3*n);  \\ closure
\\ 
\\ A string function name is eval()ed to get its global value.  A local my()
\\ function must be given by its closure.  
\\
\\ If you don't have the .internal.txt for the OFFSET but do have the
\\ stripped file for DATA, then the offset can be given explicitly in the
\\ flags parameter,
\\
\\     OEIS_check_func("A000290", "sqr", ['offset,0]);
\\
\\ If the checks pass then the return is 1, if not then print a report and
\\ error() out.
\\
\\
\\ B-File Checks
\\ -------------
\\
\\ Sequence terms from a b-file can be checked by symbol 'bfile in the flags
\\ parameter.  Remember this is after the (optional) function name argument,
\\ so is a double comma if the default suits there,\\
\\
\\     A016742(n) = (2*n)^2;
\\     OEIS_check_func("A016742",,'bfile);
\\     
\\
\\ Checks Verbosity
\\ ----------------
\\
\\ The OEIS_check_verbose global variable can be set to have a short print
\\ of each check,
\\ 
\\     OEIS_check_verbose = 1
\\     
\\ The exact form of this is unspecified, but is intended as a summary of
\\ checks made.
\\
\\ 
\\ Vector Check
\\ ------------
\\
\\ OEIS_check_vecfunc() makes the same checks as above but func() takes a
\\ number of terms to calculate and returns a vector
\\ 
\\     A000290_vector(len) = \
\\       my(t=0); vector(len,i, t+=2*i-1);
\\     OEIS_check_func("A000290_vector")
\\
\\ Sometimes calculating a vector of terms is more efficient (when they're
\\ all wanted), or sometimes the only practical way to calculate a given
\\ term is to calculat everything up to there.
\\ 
\\ Flags can be symbol 'bfile to check from a b-file as above.
\\
\\
\\ Predicate Check
\\ ---------------
\\
\\ OEIS_check_predfunc() checks a predicate function against sequence DATA
\\ or b-file.
\\ 
\\     iscube(x) = ispower(x,3);
\\     OEIS_check_predfunc("A000578","iscube");
\\     
\\     bool = iscube(x) for x = 0 .. DATA max
\\     must return 0 or 1
\\     
\\ If the sequence includes negative terms then x starts from the smallest
\\ term.  In all cases x is limited to
\\ 
\\     -100000 <= x <= 100000
\\     
\\ to avoid unfeasibly large ranges.  This limit is hard coded currently but
\\ will become a flags parameter.
\\ 
\\ The anum parameter can be the predicate function, of the form
\\ "is_A999999_optional_blah".
\\ 
\\     is_A000578_cube(x) = ispower(x,3);
\\     OEIS_check_predfunc("is_A000578_cube");
\\
\\ Note that sequences which are already characteristic functions are
\\ checked as a plain function,
\\
\\     OEIS_check_func("A010057","iscube",['offset,0])
\\
\\ Rows Check
\\ ----------
\\
\\ OEIS_check_rowfunc() checks a row-making function against sequence DATA
\\ or b-file.
\\ 
\\     row(n) = if(n==0,[0],binary(n))
\\     OEIS_check_rowfunc("A000578","row");
\\     
\\     bool = iscube(x) for x = 0 .. DATA max
\\     must return 0 or 1
\\     

\\-----------------------------------------------------------------------------
\\ with an idea to print nice length lines
\\
OEIS_print_STU(vec) =
{
  my(STU=['S,'T,'U],
    upto=1);
  for(i=1,#STU,
     my(join=Str("%",STU[i]," "),
        str="");
     while(1,
           if(upto>#vec,break(2));
           my(s=Str(join,vec[upto]));
           if(length(str)+length(s)>70, break);
           str=concat(str,s);
           upto++;
           join=",");
     if(i!=#STU,str=concat(str,","));
     print(str));
}

OEIS_antidiagonals_of_func(len,func, row_start=1,col_start=1,down=0) =
{
  my(v=vector(len),upto=0);
  for(d=0,len,
     for(o=0,d,
        if(upto>=#v,break(2));
        my(row=d-o, col=o);
        if(down,[row,col]=[col,row]);
        v[upto++]=func(row+row_start,col+col_start)));
  v;
}

\\-----------------------------------------------------------------------------
\\ Checks

\\ OEIS_INTERNAL_leading_anum(s) returns a string which is an A-number at
\\ the start of string or symbol s.
\\
my(d=Set(Vecsmall("A0123456789")), \
   is=Vecsmall("is_")); \
OEIS_INTERNAL_leading_anum(s) =
{
  s=Str(s);
  my(v=Vecsmall(s));

  \\ remove leading "is_" or "is"
  if(#v>=3 && v[1..3] == is, v=v[4..#v],
     #v>=2 && v[1..2] == is, v=v[3..#v]);

  \\ keep as much as A99... 
  for(i=1,#v,
     if(!setsearch(d,v[i]),
        v=v[1..i-1]; break));
  Strchr(v);
}
  
OEIS_check_func(anum,func=anum,flag=0) =
{
  anum=OEIS_INTERNAL_leading_anum(anum);
  my(offset='none,
     data_flag=0,
     pred=0);
  if(type(flag)=="t_POL", flag=[flag]);
  for(i=1,#flag,
     if(flag[i]=='bfile,  data_flag='bfile,
        flag[i]=='offset, offset=flag[i++],
        error("Unrecognised option at i=",i,": ",flag[i])));

  my(funcname=func);
  if(type(func)=="t_STR", func=eval(func));
  if(type(func)!="t_CLOSURE",
     error("No such function: ",funcname));
  if(type(funcname)=="t_CLOSURE",
     funcname="(anonymous function)");

  my(data);
  if(offset=='none && !pred,
     [offset,data]=OEIS_offset_and_data(anum,data_flag),
     data=OEIS_data(anum,data_flag));
  if(OEIS_check_verbose,
     print("check ",anum," (",#data," terms) vs ",funcname));

  my(got);
  my(upto=0,
     n=offset,
     t=0,
     bad=0);
  got=vector(#data);
  while(upto<#got,
        my(g=func(n));
        if(type(g) != "t_INT",
           if(bad++<5,
              print(anum," check want t_INT got type ",type(g)," "g)));
        if(t && type(g)!=t,
           error("Inconsistent func returns ",t," and ",type(g)));
        if(type(g)=="t_VEC",
           for(i=1,#g,
              if(upto>=#got,break(2));
              got[upto++]=g[i]),
           got[upto++] = g);
        n++;);;
  if(bad || got!=data,
     print(anum," by function ",funcname);
     print("got  ",got);
     print("want ",data);
     print("diff ",iferr(got-data,e,errname(e)));
     error(anum," check failed"));
  1;
}
{
addhelp(OEIS_check_func,
"OEIS_check_func(anum,func=anum,flag=0)
Check \"anum\" DATA against function func() and return 1 if good or throw an
error() if bad (with soem diagnostic prints too).

func is called \"term = func(n)\" for each index n in the DATA range.
See notes in comments at the start of OEIS-data.gp for more.");
}

\\-----------------------------------------------------------------------------
\\ Row Check

OEIS_check_rowfunc(anum,func=anum,flags=0) =
{
  anum=OEIS_INTERNAL_leading_anum(anum);
  my(funcname=func);
  if(type(func)=="t_STR", func=eval(func));
  if(type(func)!="t_CLOSURE",
     error("No such function: ",funcname));
  if(type(funcname)=="t_CLOSURE",
     funcname="(anonymous function)");

  my(offset,data);
  [offset,data]=OEIS_offset_and_data(anum,flags);
  if(OEIS_check_verbose,
     print("check ",anum," (",#data," terms) vs rows ",funcname));

  my(n=offset,
     got_all=[],
     i=1,  \\ data[i..] is next row
     bad=0);
  while(i<=#data,
        \\ print("at i="i" n="n);
        my(got=func(n));
        if(type(got) != "t_VEC",
           print(anum," by row func ",funcname);
           error("row n="n" want t_VEC got type ",type(got)));
        for(j=1,#got,
           if(type(got[j]) != "t_INT",
              print(anum," by row func ",funcname);
              error("row n="n" want t_INT got["j"] type ",type(got[j])," ",got[j])));

        if(#got!=0,
           got_all=concat(got_all,got);
           my(end=min(i+#got-1,#data));
           i<=end || error();
           my(want=data[i..end]);
           if(!bad && !(got[1..#want] === want),
              bad = 1;
              print(anum," by row func ",funcname);
              print("row n="n);
              print("want ",want);
              print("got  ",got);
              print("diff  ",got[1..#want]-want);
              if(got==want,
                 print("due to got data types not integers")));
           i = end+1);
        n++);
  if(bad,
     print("got all  ",got_all[1..#want]);
     print("want     ",want);
     error(anum," check failed"));
  1;
}

\\-----------------------------------------------------------------------------
\\ Vec Check

OEIS_check_vecfunc(anum,func=anum,flags=0) =
{
  anum=OEIS_INTERNAL_leading_anum(anum);
  my(funcname=func);
  if(type(func)=="t_STR", func=eval(func));
  if(type(func)!="t_CLOSURE",
     error("No such function: ",funcname));
  if(type(funcname)=="t_CLOSURE",
     funcname="(anonymous function)");

  my(data=OEIS_data(anum,flags));
  if(OEIS_check_verbose,
     print("check ",anum," (",#data," terms) vs vec ",funcname));

  my(got=func(#data),
     bad=0);
  for(j=1,#got,
     if(type(got[j]) != "t_INT",
        if(bad++<5,
           print(anum," by vec func ",funcname);
           print("want t_INT got vec["j"] type ",type(got[j])," ",got[j]))));

  if(got!=data,
     print("got  ",got);
     print("want ",data);
     error(anum," check failed"));
  1;
}

\\-----------------------------------------------------------------------------
\\ Predicate Check

OEIS_check_predfunc(anum,func=anum,flags=0) =
{
  anum=OEIS_INTERNAL_leading_anum(anum);
  my(funcname=func);
  if(type(func)=="t_STR", func=eval(func));
  if(type(funcname)=="t_CLOSURE",
     funcname="(anonymous function)");

  my(data=OEIS_data(anum,flags));
  if(OEIS_check_verbose,
     print("check ",anum," (",#data," terms) vs pred ",funcname));

  my(limit=100000);
  data=Set(data);
  data = select(t->abs(t) < limit, data);
  my(di=1,
     bad=0);
  for(i=min(0,data[1]), data[#data],
     my(want=(i==data[di]),
        got=func(i));
     if(! (got === want),
        if(bad++ > 10,
           print("and more ...");
           error("predicate wrong: ",funcname));
        print("pred("i") want "want" got ",got,
              if(got==want,Str(" (type ",type(got)")"),"")));
     if(want, di++));
  1;
}
{
addhelp(OEIS_check_predfunc,
"OEIS_check_predfunc(anum,func,flags=0)

Check \"anum\" DATA against predicate function func() and return 1 if good
or throw an error() if bad (with soem diagnostic prints too).

func is called \"bool = func(x)\" and must return 1 if x is a term of the
sequence, or return 0 if not.  See further notes in comments at the start of
OEIS-data.gp.");
}

\\-----------------------------------------------------------------------------
